// Listing9-7.S
//
// 32-bit ARM code for the
// decimal string to numeric conversion.
//
// Written to run on a Raspberry Pi.

#include    "ao32aa.inc"

// Number of iterations for testing:

#define iterations 100000000

// The following are required to use ARM V7&8 instructions
// on Raspberry Pi devices (3, 4, & 5):

            .arch   armv8-a

// Manifest constant for tab character:

#define tab 0x9

        .data
    
// Data to convert:
    
in0:    .asciz  "0"
in1:    .asciz  "1"
in2:    .asciz  "22"
in3:    .asciz  "333"
in4:    .asciz  "4444"
in4a:   .asciz  "1234"
in4b:   .asciz  "5678"
in5:    .asciz  "55555"
in5a:   .asciz  "12345"
in5b:   .asciz  "67890"
in6:    .asciz  "666666"
in7:    .asciz  "7777777"
in8:    .asciz  "88888888"
in9:    .asciz  "999_999_999"
in10:   .asciz  "4_294_967_295"
in11:   .asciz  " ;:,\t 12345678 "
in12:   .asciz  "1_999_999_999"

ine:    .asciz  ""
inf:    .asciz  "*1234"
ing:    .asciz  "12*34"
inh:    .asciz  "12345678900"

// Pointers to strings, so asmMain can process
// all of them in a loop:
        
strPtrs:
        .word   in0, in1, in2, in3
        .word   in4, in4a, in4b
        .word   in5, in5a, in5b
        .word   in6, in7
        .word   in8, in9, in10, in11, in12
        .set    inCnt, (.-strPtrs)/4    //Number of strings

        .text
        .align  2

        #include    "print.inc"        

// getTitle returns the address of a string in
// the R0 register. The c.cpp program will print
// this string before and after calling asmMain.

        .text
        .align  2
        
// Here's the title string that c.cpp will print:

title:  wastr  "Listing9-7"
        
        proc    getTitle, public
        adr     r0, title
        bx      lr
        endp    getTitle

////////////////////////////////////////////////////
//
// bufToU32
//
// Decimal string to 32-bit integer conversion.
// This code allows underscores to appear within a
// decimal string (but underscores cannot be the first
// character of a decimal string). This code also
// allows an underscore at the end of the string;
// if you consider this to be an error, modify
// the code at label "delimiter" to test for an
// underscore in the previous character position.
//
// On entry:
//
//  R0-     Pointer to buffer holding string.
//
// On exit-
//
//  R0-     Points at the first character beyond
//          the converted string, if conversion
//          was successful.
//
//  R1-     Contains the converted value if the
//          conversion was successful.
//
//  Carry-  Clear if conversion was successful.
//          Set if unsuccessful. If unsuccessful
//          conversion, then R0 will point at
//          the character the conversion failed
//          on and R1 will contain one of the
//          following error codes:
//
//          -1: First (non-WS) character was not
//              a valid decimal digit.
//
//          -2: Overflow (value did not fit into
//              32 bits).
//
//          -3: Last character was not a valid
//              delimiter character (this could
//              be considered a "soft" error).
//
//  This function will skip any leading whitespace
// (spaces or tabs) appearing at the beginning
// of the string.
        
#define delimCode 11
#define underscore 10
#define illChar 0xff

    .macro  entry
    push    {r2-r6, lr}
    .endm
    
    .macro  rtn
    pop     {r2-r6, pc}
    .endm

    proc    bufToU32, thumb, public
    entry
    
    
    ldr     r3, =42946729   // For overflow check
    ldr     r4, =classTbl   // For char classification
    mov     r6, #10         // For radix multiplication
    
    // Skip any leading whitespace, comma, semicolon,
    // or colon characters in the string:
    
skipDelim:
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]    // Get classification
    cmp     r2, #0          // Check for End of String
    beq     badFirst
    cmp     r5, #delimCode  // Is it a delimiter?
    beq     skipDelim
    
    // First non-delimiter character must
    // be a valid decimal digit:
    
    mov     r1, r5
    blo     nextChar        // Decimal digit
            
    // If the first digit is bad (including underscore), 
    // return -1:
    
badFirst:
    sub     r0, r0, #1          // Point back at bad char
    mov     r1, #-1
    cmp     r1, r1              // Sets carry flag
    rtn 

    // Process decimal digits.
    // Unroll the character conversion loop eight times 
    // to improve performance.
    
    .macro  doDec1
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]
    cmp     r5, #underscore
    beq     nextChar
    bhi     delimiter
    muls    r1, r1, r6 
    adds    r1, r1, r5
    .endm

    .macro  doDec2
    ldrb    r2, [r0], #1
    ldrb    r5, [r4, r2]
    cmp     r5, #underscore
    beq     nextChar
    bhi     delimiter
    cmp     r1, r3
    bhs     overflow
    muls    r1, r1, r6 
    adds    r1, r1, r5
	bcs		overflow
    .endm

    
nextChar:
    doDec1                  // Assume no more than 8 digits.
    doDec1                  // If 8 digits or less, overflow
    doDec1                  // will not occur. Note that we've
    doDec1                  // already processed the first digit.
    doDec1
    doDec1
    doDec1
    
    // If there are more than 8 digits, we need to check for
    // overflow (overflow is not guaranteed, there could be
    // leading zeros):
    
couldOverflow:
    doDec2
    doDec2
    doDec2
    doDec2
    doDec2
    doDec2
    doDec2
    doDec2
    b       couldOverflow       // If more than 16 digits.
            
    // Transfer here when encountering a legitimate
    // delimiter character (valid end of decimal string).
    
delimiter:
    sub     r0, r0, #1          // Point back at delimiter
    cmp     r5, #delimCode
    bne     illegal
    adds    r0, r0, #0          // Clears carry flag
    rtn

    // If the first non-decimal-digit character
    // is bad, return -3:
    
illegal:
    mov     r1, #-3
    cmp     r1, r1              // Sets carry flag
    rtn

    
overflow:
    mov     r1, #-2
    cmp     r1, r1              // Sets carry flag
    rtn
    endp    bufToU32

    // Classify each character:
    //
    //  0 to 9-     decimal character
    //  0x10-       underscore
    //  0x11-       delimiter character
    //  0xff-       illegal character
    
classTbl:
    .byte   delimCode   // 0x00
    .byte   illChar     // 0x01
    .byte   illChar     // 0x02
    .byte   illChar     // 0x03
    .byte   illChar     // 0x04
    .byte   illChar     // 0x05
    .byte   illChar     // 0x06
    .byte   illChar     // 0x07
    .byte   illChar     // 0x08
    .byte   delimCode   // 0x09
    .byte   delimCode   // 0x0a
    .byte   illChar     // 0x0b
    .byte   illChar     // 0x0c
    .byte   delimCode   // 0x0d
    .byte   illChar     // 0x0e
    .byte   illChar     // 0x0f
    .byte   illChar     // 0x10
    .byte   illChar     // 0x11
    .byte   illChar     // 0x12
    .byte   illChar     // 0x13
    .byte   illChar     // 0x14
    .byte   illChar     // 0x15
    .byte   illChar     // 0x16
    .byte   illChar     // 0x17
    .byte   illChar     // 0x18
    .byte   illChar     // 0x19
    .byte   illChar     // 0x1a
    .byte   illChar     // 0x1b
    .byte   illChar     // 0x1c
    .byte   illChar     // 0x1d
    .byte   illChar     // 0x1e
    .byte   illChar     // 0x1f
    .byte   delimCode   // ' '
    .byte   illChar     // '!'
    .byte   illChar     // '"'
    .byte   illChar     // '#'
    .byte   illChar     // '$'
    .byte   illChar     // '%'
    .byte   illChar     // '&'
    .byte   illChar     // '''
    .byte   illChar     // '('
    .byte   illChar     // ')'
    .byte   illChar     // '*'
    .byte   illChar     // '+'
    .byte   delimCode   // ','
    .byte   illChar     // '-'
    .byte   illChar     // '.'
    .byte   illChar     // '/'
    .byte   0           // '0'
    .byte   1           // '1'
    .byte   2           // '2'
    .byte   3           // '3'
    .byte   4           // '4'
    .byte   5           // '5'
    .byte   6           // '6'
    .byte   7           // '7'
    .byte   8           // '8'
    .byte   9           // '9'
    .byte   delimCode   // ':'
    .byte   delimCode   // ';'
    .byte   illChar     // '<'
    .byte   illChar     // '='
    .byte   illChar     // '>'
    .byte   illChar     // '?'
    .byte   illChar     // '@'
    .byte   illChar     // 'A'
    .byte   illChar     // 'B'
    .byte   illChar     // 'C'
    .byte   illChar     // 'D'
    .byte   illChar     // 'E'
    .byte   illChar     // 'F'
    .byte   illChar     // 'G'
    .byte   illChar     // 'H'
    .byte   illChar     // 'I'
    .byte   illChar     // 'J'
    .byte   illChar     // 'K'
    .byte   illChar     // 'L'
    .byte   illChar     // 'M'
    .byte   illChar     // 'N'
    .byte   illChar     // 'O'
    .byte   illChar     // 'P'
    .byte   illChar     // 'Q'
    .byte   illChar     // 'R'
    .byte   illChar     // 'S'
    .byte   illChar     // 'T'
    .byte   illChar     // 'U'
    .byte   illChar     // 'V'
    .byte   illChar     // 'W'
    .byte   illChar     // 'X'
    .byte   illChar     // 'Y'
    .byte   illChar     // 'Z'
    .byte   illChar     // '['
    .byte   illChar     // '\'
    .byte   illChar     // ']'
    .byte   illChar     // '^'
    .byte   underscore  // '_'
    .byte   illChar     // '`'
    .byte   illChar     // 'a'
    .byte   illChar     // 'b'
    .byte   illChar     // 'c'
    .byte   illChar     // 'd'
    .byte   illChar     // 'e'
    .byte   illChar     // 'f'
    .byte   illChar     // 'g'
    .byte   illChar     // 'h'
    .byte   illChar     // 'i'
    .byte   illChar     // 'j'
    .byte   illChar     // 'k'
    .byte   illChar     // 'l'
    .byte   illChar     // 'm'
    .byte   illChar     // 'n'
    .byte   illChar     // 'o'
    .byte   illChar     // 'p'
    .byte   illChar     // 'q'
    .byte   illChar     // 'r'
    .byte   illChar     // 's'
    .byte   illChar     // 't'
    .byte   illChar     // 'u'
    .byte   illChar     // 'v'
    .byte   illChar     // 'w'
    .byte   illChar     // 'x'
    .byte   illChar     // 'y'
    .byte   illChar     // 'z'
    .byte   illChar     // '{'
    .byte   illChar     // '|'
    .byte   illChar     // '}'
    .byte   illChar     // '~'
    .byte   illChar     // del
    .space  128, illChar    // 0x80 to 0xFF

/////////////////////////////////////////////////////////////////////
//
// strToH32
//
// Converts the string pointed at by R0 to a numeric value
// and returns that numeric value in R1. This function
// preserves R0 (leaves it pointing at the beginning of the
// string buffer).

    proc    strToU32, thumb, public
    push    {r0, lr}
    bl      bufToU32
    pop     {r0, pc}
    endp    strToU32
            
/////////////////////////////////////////////////////////////////////
//              
// Here's the assembly language main program:
        
            proc    asmMain, public
            push    {r0-r8, lr}

// Timing code:

    ldr     r8, =iterations
timingLp:

    ldr     r7, =strPtrs
    mov     r6, #inCnt
innerLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1:
    
    bl      bufToU32
            
    // Repeat for each sample input:
      
    subs    r6, #1
    bne     innerLp            

    // Repeat a large number of times in order
    // to time the code:
    
    subs    r8, r8, #1
    bne     timingLp


// Output code for testing:

    ldr     r7, =strPtrs
    mov     r6, #inCnt
convLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1 and output
    // that value:
    
    bl      strToU32
    bcc     goodConv
    
    bl      print
    wastr   "Conversion error processing \"%s\", code = %ld\n"
    b       nextConv

goodConv:
    bl      print
    wastr   "Converted '%s' to %lu\n"
          
nextConv:
    subs    r6, #1
    bne     convLp
    
// Check for expected errors:
//
// Check empty string on input:

    ldr     r0, =ine
    bl      strToU32
    bcs     goodErr1
    bl      print
    wastr   "Expected error for empty string input\n"
    b       tryInf
    
goodErr1:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check "*1234" on input:
    
tryInf:     
    ldr     r0, =inf
    bl      strToU32
    bcs     goodErr2
    bl      print
    wastr   "Expected illegal 1st char for \"%s\"\n"
    b       tryIng
    
goodErr2:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check "12*34" on input:
    
tryIng:     
    ldr     r0, =ing
    bl      strToU32
    bcs     goodErr3
    bl      print
    wastr   "Expected illegal 1st char for \"%s\"n"
    b       tryInh
    
goodErr3:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"

// Check for overflow:
    
tryInh:     
    ldr     r0, =inh
    bl      strToU32
    bcs     goodErr4
    bl      print
    wastr   "Expected illegal 1st char for \"%s\"n"
    b       allDone
    
goodErr4:   
    bl      print
    wastr   "Processing \"%s\" produced error code %d\n"
                
allDone:             
    pop     {r0-r8, pc} 
    endp    asmMain


